// midi.hpp

#define true	1
#define false	0

typedef unsigned char 	byte;
typedef unsigned int  	word;
typedef unsigned long 	dword;
typedef char		  	bool;

typedef byte huge*		pbyte;
typedef word huge*		pword;
typedef dword huge*		pdword;
typedef bool huge*		pbool;

inline void swapbytes(dword& dw);
inline void swapbytes(word& w);
dword       readvarlen(pbyte& source, long& bytes_read);

class midiChunk	{
	public:
		midiChunk();
		~midiChunk();
		void read(FILE*);
		const char* type() {return id;};
		byte huge& operator[](dword idx) {return (idx<len ? data[idx] : data[len-1]);};
		dword size() {return len;};
		virtual bool verify_type() = 0;
	protected:
		byte huge* data;
		dword len;
		char  id[5];
};

class midiHeader : public midiChunk {
	public:
		word format();
		word ntracks();
		word division();
		bool verify_type();
};

class midiEvent {
	public:
		midiEvent(pbyte& raw, long& bytes_read, byte run_stat);// parses raw midi data
		midiEvent() : delta(-1) {};
		dword delta;
		bool ignore_me; // for sysex msgs, unimportant meta-events and stuff
		byte note;
		byte volume;
		byte cmd;
		byte chan;
		byte running;
};

class midiTrack : public midiChunk {
	friend class midiFile;
	public:
		midiTrack() {cur_pos=0; cur_time=0; running=0;};
		bool verify_type();
		midiEvent* nextEvent(); // 0 if no more events left in track
		midiEvent* curEvent();
		dword time() {return cur_time;};
		bool end();
		void reset();
	private:
		pbyte cur_pos;
		long bytes_read;
		midiEvent cur_event;
		dword cur_time;
		byte running;
};

class midiFile {
	public:
		midiFile() {_track=0;};
		~midiFile() {if(_track!=0) delete[]_track;};
		void read(FILE* file);
		bool finished();
        int nexttrack();
		midiHeader* header() {return &_header;};
		midiTrack* track(int);
		void reset();
	protected:
		midiHeader _header;
		midiTrack *_track;
};
