// Converts MIDI files to XM files

#include <stdio.h>
#include <iostream.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "midi.hpp"

//#define SHAREWARE

#ifdef  SHAREWARE
#	  define MAXCHAN 	10
#else
#	  define MAXCHAN	32
#endif

#define TEMPO   	3
#define BPM			120
#define PATTERNLEN	0x40
#define VERSION		"1.10"
#define PASS1		1
#define PASS2		2

static byte note[MAXCHAN]; // note being played by each channel (255=none, 254=hold)
static byte chanchan[MAXCHAN]; // which MIDI chan is assigned to each XM chan
static byte instr[MAXCHAN]; // instruments currently in use by each channel
static byte volume[MAXCHAN]; // volume of current note
static byte delay[MAXCHAN]; // delay of current note, in ticks
static word pattern_len; // length of current pattern
static int npatterns,row;
static pbyte xmbuf;
static int factor;
static dword tempo;
static bool tempochg; // specifies whether tempo has changed
static byte speed; // current speed
static int trans; // transpose value
static int nchan; // maximum number of channels used simultaneously

class InstrList {
	public:
		InstrList();
		byte getnew(byte oldnum);
		byte getold(byte newnum);
		int ninstr() {return highest;};
		void write(FILE* file);
	private:
		byte instr[128];
		byte alias[128];
		char name[128][80];
		byte highest;
} instrlist;

void writeheader(FILE* xm, midiFile& midi, const char* name);
int  findfreechan(int pass);
void write_row(FILE*);
void newpattern(FILE*);
void write_new_tempo();
void getfilename(char* fname, int n);

int main()
{
	char filename[80],dfname[80];
	FILE *f_midi,*f_xm;
	int i,j;
	midiFile midi;
	midiEvent *event;
	dword time=0;

	memset(chanchan,0,MAXCHAN);
	memset(note,255,MAXCHAN);
	xmbuf = 0;

	clrscr();
	cout<<"MidiMix "<<VERSION<<endl
		<<"Converts MIDI files to FastTracker XM files"<<endl<<endl
		<<"Enter MIDI input file name:"<<endl;
	cin>>filename;
	if(strstr(filename,".")==0)
		strcat(filename,".mid");
	f_midi=fopen(filename,"rb");
	if(f_midi==0) {
		cout<<"File not found - "<<filename<<endl;
		exit(1);
	}
	strcpy(dfname,filename);
	strcpy(strstr(dfname,"."),".xm");
	cout<<"Enter XM output file name ["<<dfname<<"]:"<<endl;
	fgets(filename,80,stdin);
	*strstr(filename,"\n")=0;
	if(filename[0]==0)
		strcpy(filename,dfname);
	if(strstr(filename,".")==0)
		strcat(filename,".xm");
	f_xm=fopen(filename,"rb");
	if(f_xm!=0) {
		cout<<"File "<<filename<<" already exists. Overwrite? (Y/N) ";
		if(toupper(getche())=='Y')
			fclose(f_xm);
		else {
			cout<<"Program aborted.\n";
			exit(1);
		}
	}
	cout<<endl;
	f_xm=fopen(filename,"wb");
	if(f_xm==0) {
		cout<<"Can't open file "<<filename<<".\nCheck disk space and write permission."<<endl;
		exit(1);
	}
	cout<<"Enter number of rows per quarter note (8 recommended): ";
	cin>>factor;
	if(factor<=0) {
		cout<<"Illegal number entered - 8 channels will be used.\n";
		factor=8;
	}
	if(factor>32) {
		cout<<"Illegal value entered - maximum value of 32 will be used.\n";
		factor=32;
	}
	cout<<"Enter number of octaves to transpose (e.g. -1 for 1 octave down): ";
	cin>>trans;
	if(trans<-4) trans=-4;
	if(trans>4) trans=4;
	if(trans==0)
		cout<<"No transposition will be made.\n";
	else if(trans>0)
		cout<<"Transposing "<<trans<<" octave(s) up.\n";
	else
		cout<<"Transposing "<<-trans<<" octave(s) down.\n";
	midi.read(f_midi);
	fclose(f_midi);

	clrscr();
	if(midi.header()->format()!=0 && midi.header()->format()!=1) {
		cout<<"Invalid MIDI file format."<<endl
			<<"MidiMix does not support time-asynchronus MIDI tracks."<<endl;
		exit(1);
	}
	if((midi.header()->division()&0x8000) != 0) {
		cout<<"Invalid MIDI file format."<<endl
			<<"MidiMix does not support absolute timeframe MIDI files."<<endl;
		exit(1);
	}

	cout<<"Processing MIDI file - Pass 1...\n";
	while(!midi.finished()) {
		i = midi.nexttrack();
		for( ; time<midi.track(i)->time() ; time+=midi.header()->division()/factor)
			; // null loop
		event = midi.track(i)->curEvent();
		switch(event->cmd) {
			case 0xc0: // program change
				instr[event->chan] = event->note;
				break;
			case 0x90: // note on
				j = findfreechan(PASS1);
				note[j] = event->note;
				chanchan[j] = event->chan;
				volume[j] = event->volume;
				delay[j] = (midi.track(i)->time()-time+midi.header()->division()/factor)*speed*factor/midi.header()->division();
				if(delay[j]==speed && delay[j]!=0) delay[j]--;
				if(delay[j]>=speed && delay[j]!=0) {
					cout<<"Error in delay calculation!"<<endl;
					abort();
				}
				break;
			case 0x80: // note off
				for(j=0; j<nchan; j++)
					if(note[j]==event->note && chanchan[j]==event->chan)
						break;
				if(j==nchan) {
					cout<<"Warning on Note "<<(int)event->note<<", Channel "
						<<(int)event->chan<<", Track "<<i<<", Time "<<time
						<<endl;
					break; // not found; just forget it
				}
				volume[j]=0;
				note[j]=255;
				chanchan[j]=255;
				break;
			case 0x51: // tempo change
				tempo = ((dword)event->note<<16) | ((dword)event->chan<<8)
						| ((dword)event->volume);
				tempochg = true;
				break;
			default:
				cout<<"Unknown command.\n";
				abort();
		}
		midi.track(i)->nextEvent();
	}

	if(nchan%2==1) nchan++;
	if(nchan>MAXCHAN) {
		cout<<"This MIDI file requires "<<nchan<<" channels.\n";
#		ifdef SHAREWARE
		cout<<"This version supports up to "<<MAXCHAN<<" channels."<<endl
			<<"The registered version supports up to 32 channels."<<endl
			<<"For registration information, read the documentation or send mail to"<<endl
			<<"zvikabh@tx.technion.ac.il"<<endl;
#		else
		cout<<"The FastTracker XM format only allows up to 32 channels, but the specified"<<endl
			<<"MIDI file uses more than that. This MIDI file cannot be converted into"<<endl
			<<"XM format. This is a limitation in the XM format, not in MidiMix."<<endl;
#		endif
		exit(1);
	}

	cout<<"Using "<<nchan<<" channels in XM file.\n";
	writeheader(f_xm,midi,filename);
	memset(chanchan,0,MAXCHAN);
	memset(note,255,MAXCHAN);
	cout<<"Processing MIDI file - Pass 2...\n";
	midi.reset();
	time=0;
	while(!midi.finished()) {
		i = midi.nexttrack();
		for( ; time<midi.track(i)->time() ; time+=midi.header()->division()/factor)
			write_row(f_xm);
		event = midi.track(i)->curEvent();
		switch(event->cmd) {
			case 0xc0: // program change
				instr[event->chan] = event->note;
				break;
			case 0x90: // note on
				j = findfreechan(PASS2);
				note[j] = event->note;
				chanchan[j] = event->chan;
				volume[j] = event->volume;
				delay[j] = (midi.track(i)->time()-time+midi.header()->division()/factor)*speed*factor/midi.header()->division();
				if(delay[j]==speed && delay[j]!=0) delay[j]--;
				if(delay[j]>=speed && delay[j]!=0) {
					cout<<"Error in delay calculation!"<<endl;
					abort();
				}
				break;
			case 0x80: // note off
				for(j=0; j<nchan; j++)
					if(note[j]==event->note && chanchan[j]==event->chan)
						break;
				if(j==nchan) {
					break; // not found; just forget it
				}
				volume[j]=0;
				note[j]=255;
				chanchan[j]=255;
				break;
			case 0x51: // tempo change
				tempo = ((dword)event->note<<16) | ((dword)event->chan<<8)
						| ((dword)event->volume);
				tempochg = true;
				break;
			default:
				cout<<"Unknown command.\n";
				abort();
		}
		midi.track(i)->nextEvent();
	}
	do
		write_row(f_xm);
	while(row!=0);

	fclose(f_xm);
	f_xm=fopen(filename,"r+b");
	fseek(f_xm,64,0);
	fputc((byte)npatterns,f_xm);
	fseek(f_xm,70,0);
	fputc((byte)(npatterns-1),f_xm);
	fseek(f_xm,72,0);
	fputc((byte)instrlist.ninstr()+12,f_xm);
	fseek(f_xm,80,0);
	for(i=0; i<npatterns; i++)
		fputc(i,f_xm);
	fseek(f_xm,0,2);
	cout << npatterns << " patterns written.                   \n";
	instrlist.write(f_xm);
	delete [] xmbuf;
	return 0;
}

#pragma argsused
void writeheader(FILE* xm, midiFile& midi, const char* name) {
	char header[80+256+9];
	memset(header,0,80+256);
	sprintf(header,"Extended Module: %s",name);
	header[37] = 0x1a;
	sprintf(header+38,"MidiMix %s",VERSION);
	*(word*)(header+58) = 0x0104;
	*(dword*)(header+60) = 0x114;
	*(word*)(header+68) = nchan;
	*(word*)(header+74) = 1;
	*(word*)(header+76) = TEMPO;
	*(word*)(header+78) = BPM;
	fwrite(header,1,80+256,xm);
	newpattern(xm);
}

void newpattern(FILE* xm) {
	if(xmbuf!=0) {
		*(dword*)(xmbuf) = 9;
		xmbuf[4] = 0;
		*(word*)(xmbuf+5) = PATTERNLEN;
		*(word*)(xmbuf+7) = pattern_len-9;
		fwrite(xmbuf,1,pattern_len,xm);
		row=0;
	} else {
		xmbuf = new char[5*MAXCHAN*PATTERNLEN+100];
		if(xmbuf==0) {
			cout<<"Insufficient memory.\n";
			exit(1);
		}
		memset(xmbuf,0x80,5*nchan*PATTERNLEN);
		row=-1;
	}
	npatterns++;
	cout<<"Writing pattern #"<<npatterns-1<<"...\r";
	pattern_len=9;
}

int findfreechan(int pass) {
	int i;
	for(i=0; i<nchan; i++)
		if(note[i]==255) return i;
	if(pass==PASS1) {
		if(nchan>=MAXCHAN) {
#			ifdef SHAREWARE
				cout<<"The MIDI file you are attempting to convert contains too many channels."<<endl
					<<"This version supports up to "<<MAXCHAN<<" channels."<<endl
					<<"The registered version supports up to 32 channels."<<endl
					<<"For registration information, read the documentation or send mail to"<<endl
					<<"zvikabh@tx.technion.ac.il"<<endl;
#			else
				cout<<"The FastTracker XM format only allows up to 32 channels, but the specified"<<endl
					<<"MIDI file uses more than that. This MIDI file cannot be converted into"<<endl
					<<"XM format. This is a limitation in the XM format, not in MidiMix."<<endl;
#			endif
		exit(1);
		}
		return nchan++;
	}
	cout<<"Error #92: This error should not occur. Please contact the author\n"
		<<"at zvikabh@tx.technion.ac.il and specify the MIDI file causing the\n"
		<<"error, and the parameters you used.\n";
	exit(1);
	return -1; // keep the compiler happy
}

void write_row(FILE* xm) {
	int i=0;
	if(++row==PATTERNLEN)
		newpattern(xm);
	if(tempochg) {
		i=2; // first 2 channels will be written by write_new_tempo
		write_new_tempo();
	}
	for(; i<nchan; i++) {
		if(volume[i]==254) {				// note hold
			xmbuf[pattern_len++] = 0x80;
		} else if(note[i]==255) {	 		// note off
			xmbuf[pattern_len++] = 0x81;
			xmbuf[pattern_len++] = 0x61;
			volume[i]=254;
		} else {					// ordinary note
			if(delay[i]==0) {
				xmbuf[pattern_len++] = 0x87;
				xmbuf[pattern_len++] = note[i]-11+12*trans;
				xmbuf[pattern_len++] = instrlist.getnew(instr[i])+1;
				xmbuf[pattern_len++] = volume[i]/2 + 0x10;
				volume[i]=254; // hold note until further notice
			} else {
				xmbuf[pattern_len++] = 0x9F;
				xmbuf[pattern_len++] = note[i]-11+12*trans;
				xmbuf[pattern_len++] = instrlist.getnew(instr[i])+1;
				xmbuf[pattern_len++] = volume[i]/2 + 0x10;
				xmbuf[pattern_len++] = 0x0e;
				xmbuf[pattern_len++] = 0xd0 | delay[i];
				volume[i]=254; // hold note until further notice
			}
		}
	}
}

byte InstrList::getnew(byte oldnum) {
	if(instr[alias[oldnum]]!=255)
		return instr[alias[oldnum]];
	return (instr[alias[oldnum]] = highest++);
}

byte InstrList::getold(byte newnum) {
	int i;
	if(newnum>=highest)
		return 255;
	for(i=0; i<128; i++)
		if(instr[i]==newnum)
			return i;
	return 255; // shouldn't happen
}

void InstrList::write(FILE* file) {
	int i,j,max;
	char buffer[0x21];
	char table[0x60];
	char *pch;
	pbyte buf2;
	FILE* source;

#	ifdef SHAREWARE
	const char blurb[12][34] = {
		"----------------------",
		"Converted from MIDI by",
		"MidiMix MID2XM Convert",
		"Shareware Version 1.1 ",
		"Programming:          ",
		"Zvika Ben-Haim        ",
		"   (a.k.a. Slarti)    ",
		"mailto:zvikabh@tx.    ",
		"   technion.ac.il     ",
		"http://www.technion.  ",
		"   ac.il/~zvikabh/    ",
		"----------------------" };
#	else
	const char blurb[12][34] = {
		"----------------------",
		"Converted from MIDI by",
		"MidiMix MID2XM Convert",
		"Registered Version 1.1",
		"Programming:          ",
		"Zvika Ben-Haim        ",
		"   (a.k.a. Slarti)    ",
		"mailto:zvikabh@tx.    ",
		"   technion.ac.il     ",
		"http://www.technion.  ",
		"   ac.il/~zvikabh/    ",
		"----------------------" };
#	endif

	buf2 = new byte[10000];
	if(buf2==0) {
		cout<<"Not enough memory.\n";
		abort();
	}

	for(i=0; i<highest; i++) {
		cout<<"Saving instrument #"<<i+1<<" ("<<name[getold(i)]<<")...";
		getfilename(table,getold(i));
		source = fopen(table,"rb");
		if(source==0) {
			cout << " Not found, using default.";
			source = fopen("SAMPLES\\00.XI","rb");
			if(source==0) {
				cout<<"\nCannot find default sample file SAMPLES\\00.XI!\n";
				exit(1);
			}
		}
		cout<<endl;
		memset(buffer,0,0x21);
		fseek(source,0x15,0);
		fread(buffer+4,1,0x16,source);
		fseek(source,0x42,0);
		fread(table,1,0x60,source);
		for(j=max=0; j<0x60; j++)
			if(table[j]>max)
				max=table[j];
		buffer[0x1b] = max+1;
		buffer[0x1d] = 0x28;
		*(dword*)buffer = 0x107;
		fwrite(buffer,1,0x21,file);
		fwrite(table,1,0x60,file);
		fread(table,1,0x60,source);
		fwrite(table,1,0x60,file);
		fread(table,1,18,source);
		fwrite(table,1,18,file);
		fread(table,1,2,source); // skipped on purpose
		do {
			j=fread(buf2,1,10000,source);
			fwrite(buf2,1,j,file);
		} while(j==10000);
		fclose(source);
	}

	*(dword*)buffer = 0x21;
	*(byte*)(buffer+26)=0;
	*(word*)(buffer+27)=0;
	*(byte*)(buffer+29)=0x1e;
	*(byte*)(buffer+32)=0x36;
	for(i=0; i<12; i++) {
		memset(buffer+4,' ',22);
		for(j=0; blurb[i][j]!=0; j++)
			buffer[j+4]=blurb[i][j];
		fwrite(buffer,1,0x21,file);
	}
	cout<<endl;
	delete[] buf2;
}

void write_new_tempo() {
	tempochg=false;
	byte bpm;
	float linespersec = 1e6*(float)factor/(float)tempo;
	float bpmspeed = linespersec*2.5; // bpm/speed
	speed = 255/bpmspeed;
	bpm = bpmspeed*speed;
	if(speed>=0x20) {
		cout<<"Error in time frame. Try using more rows per quarter note.\n";
		exit(1);
	}
	if(bpm<0x20 || bpm==255) {
		cout<<"Internal time frame error. This error should not occur.\n"
			<<"Please send the author a copy of the offending MIDI file\n"
			<<"and the parameters used to convert it.\n"
			<<"e-mail: zvikabh@aluf.technion.ac.il\n";
		exit(1);
	}

	if(volume[0]==254) {		// note hold
		xmbuf[pattern_len++] = 0x98;
		xmbuf[pattern_len++] = 0x0f;
		xmbuf[pattern_len++] = speed;
	} else if(note[0]==255) {	// note off
		// using the volume-0 method to turn it off, because the ordinary
		// method doesn't allow to set an effect
		xmbuf[pattern_len++] = 0x9c;
		xmbuf[pattern_len++] = 0x00;
		xmbuf[pattern_len++] = 0x0f;
		xmbuf[pattern_len++] = speed;
		volume[0]=254;
	} else {					// ordinary note
		xmbuf[pattern_len++] = 0x9F;
		xmbuf[pattern_len++] = note[0]-11+trans*12;
		xmbuf[pattern_len++] = instrlist.getnew(instr[0])+1;
		xmbuf[pattern_len++] = volume[0]/2 + 0x10;
		xmbuf[pattern_len++] = 0x0f;
		xmbuf[pattern_len++] = speed;
		volume[0]=254; // hold note until further notice
	}

	if(volume[1]==254) {		// note hold
		xmbuf[pattern_len++] = 0x98;
		xmbuf[pattern_len++] = 0x0f;
		xmbuf[pattern_len++] = bpm;
	} else if(note[1]==255) {	// note off
		// using the volume-0 method to turn it off, because the ordinary
		// method doesn't allow to set an effect
		xmbuf[pattern_len++] = 0x9c;
		xmbuf[pattern_len++] = 0x00;
		xmbuf[pattern_len++] = 0x0f;
		xmbuf[pattern_len++] = bpm;
		volume[1]=254;
	} else {					// ordinary note
		xmbuf[pattern_len++] = 0x9F;
		xmbuf[pattern_len++] = note[1]-11+trans*12;
		xmbuf[pattern_len++] = instrlist.getnew(instr[1])+1;
		xmbuf[pattern_len++] = volume[1]/2 + 0x10;
		xmbuf[pattern_len++] = 0x0f;
		xmbuf[pattern_len++] = bpm;
		volume[1]=254; // hold note until further notice
	}
}

InstrList::InstrList() {
	FILE* file;
	char t[80];
	int i;

	memset(instr,255,128);
	highest=0;

	// read alias file
	file = fopen("smpalias.dat","rt");
	fgets(t,80,file);
	if(strcmp(t,"Sample Alias File - Do not edit this file!\n")!=0) {
		cout<<"Error in sample alias file SMPALIAS.DAT\n";
		abort();
	}
	for(i=0; i<128; i++)
		fscanf(file,"%d",alias+i);
	fclose(file);

	// read sample name file
	file = fopen("smpname.dat","rt");
	fgets(t,80,file);
	if(strcmp(t,"Sample Name File\n")!=0) {
		cout<<"Error in sample name file SMPNAME.DAT\n";
		abort();
	}
	for(i=0; i<128; i++) {
		fgets(name[i],80,file);
		name[i][strlen(name[i])-1]=0;
	}
	fclose(file);
}

void getfilename(char* fname, int n) {
	FILE* f = fopen("samples.dat","rb");
	char* pch;

	if(f==0) {
		cout<<"Error! Can't open file SAMPLES.DAT!\n";
		exit(2);
	}
	strcpy(fname,"SAMPLES\\");
	fseek(f,n*8,0);
	fread(fname+8,1,8,f);
	fname[16]=0;
	pch = strstr(fname+8," ");
	if(pch==0) pch=fname+16;
	strcpy(pch,".XI");
	fclose(f);
}
